import pytest
from pydantic import BaseModel, ValidationError

from mmuxer.models.condition import Condition
from mmuxer.utils import ParseException, find_likely_error_location_and_message


def test_find_likely_error_location_single_error():
    class M(BaseModel):
        a: str
        b: int

    data = {"a": "ok"}
    with pytest.raises(ValidationError) as exc_info:
        M.model_validate(data)
    location, message = find_likely_error_location_and_message(exc_info.value)
    assert location == ("b",)
    assert message == "Field required"


def test_find_likely_error_location_multiple_errors():
    class M(BaseModel):
        a: str
        b: int

    data = {}
    with pytest.raises(ValidationError) as exc_info:
        M.model_validate(data)
    location, message = find_likely_error_location_and_message(exc_info.value)
    assert location == ("b",) or location == ("a",)
    assert message == "Field required"


def test_find_likely_error_location_unions():
    data = {
        "condition": {
            "ALL": [
                {"FROM": "ok"},
                {"SUBECT": "ok"},
            ]
        }
    }

    class M(BaseModel):
        condition: Condition

    with pytest.raises(ValidationError) as exc_info:
        M.model_validate(data)

    location, message = find_likely_error_location_and_message(exc_info.value)
    assert location == ("condition", "All", "ALL", 1, "From")
    assert message == "could not parse"

    parse_exception = ParseException.from_validation_error(exc_info.value, data)
    assert parse_exception.bad_content == {"From": "Missing field"}


def test_find_likely_error_location_unions_multiple_errors():
    data = {"condition": {"ALL": [{"SUBECT": "ok"}, {"SENDER": "ok"}, {"ANY": ["test", "okok"]}]}}

    class M(BaseModel):
        condition: Condition

    with pytest.raises(ValidationError) as exc_info:
        M.model_validate(data)

    location, message = find_likely_error_location_and_message(exc_info.value)
    assert location == ("condition", "All", "ALL", 2, "Any", "ANY", 0)
    assert message == "could not parse"
    parse_exception = ParseException.from_validation_error(exc_info.value, data)
    assert parse_exception.bad_content == "Missing field"


def test_find_likely_error_location_unions_multiple_errors_bis():
    data = {"condition": {"ALL": [{"SUBECT": "ok"}, {"SENDER": "ok"}, {"ANY": [{}, "okok"]}]}}

    class M(BaseModel):
        condition: Condition

    with pytest.raises(ValidationError) as exc_info:
        M.model_validate(data)

    location, message = find_likely_error_location_and_message(exc_info.value)
    assert message == "could not parse"
    assert location == ("condition", "All", "ALL", 2, "Any", "ANY", 0)
    parse_exception = ParseException.from_validation_error(exc_info.value, data)
    assert parse_exception.bad_content == "Missing field"



from pydantic import BaseModel as PydanticBaseModel
from pydantic import ConfigDict, ValidationError

from mmuxer.utils import ParseException


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(extra="forbid")

    @classmethod
    def parse_data(cls, data):
        try:
            return cls.model_validate(data)
        except ValidationError as exc:
            try:
                parsed_exc = ParseException.from_validation_error(
                    exc=exc,
                    full_content=data,
                )
                raise parsed_exc
            except:
                raise exc



from pathlib import Path

from mmuxer.cli.run import _tidy
from mmuxer.config_state import state

config = """
rules:
  - move_to: Trash
    condition:
      SUBJECT: spam
settings:
  server: localhost
  username: u
  password: p
"""


def test_tidy(mailbox, make_message, tmp_path: Path):
    """
    GIVEN a mailbox with messages
    WHEN running _tidy
    THEN messages are tidied up
    """
    config_file = tmp_path / "config.py"
    config_file.write_text(config)
    state.load_config_file(config_file)
    state._mailbox = mailbox

    mailbox.login("u", "p")

    make_message(user="u", box="INBOX", subject="spam", content_text="content")
    make_message(user="u", box="INBOX", subject="ok", content_text="content")

    _tidy(None, False)

    messages = list(mailbox.fetch())
    assert len(messages) == 1

    mailbox.folder.set("Trash")
    messages = list(mailbox.fetch())
    assert len(messages) == 1



from mmuxer.models.action import DeleteAction, FlagAction, MoveAction, UnflagAction
from mmuxer.models.enums import Flag


def test_move_action(mailbox, make_message):

    moveAction = MoveAction(dest="new_folder")

    mailbox.login("u", "p")
    mailbox.folder.create("new_folder")
    make_message(user="u", box="INBOX", content_text="content")
    messages = list(mailbox.fetch())
    m = messages[0]
    assert len(messages) == 1
    moveAction.apply(mailbox, m, False)

    messages = list(mailbox.fetch())
    assert len(messages) == 0

    mailbox.folder.set("new_folder")
    messages = list(mailbox.fetch())
    assert len(messages) == 1
    m_moved = messages[0]
    assert m.uid == m_moved.uid


def test_delete_action(mailbox, make_message):

    deleteAction = DeleteAction()

    mailbox.login("u", "p")
    make_message(user="u", box="INBOX", content_text="content")
    messages = list(mailbox.fetch())
    m = messages[0]
    assert len(messages) == 1
    deleteAction.apply(mailbox, m, False)

    messages = list(mailbox.fetch())
    assert len(messages) == 0


def test_flag_actions(mailbox, make_message):

    flagAction = FlagAction(flag=Flag.FLAGGED)
    unflagAction = UnflagAction(flag=Flag.FLAGGED)

    mailbox.login("u", "p")
    make_message(user="u", box="INBOX", content_text="content")

    messages = list(mailbox.fetch())
    m = messages[0]
    assert len(messages) == 1
    print(m.flags)
    assert Flag.FLAGGED.name not in m.flags
    flagAction.apply(mailbox, m, False)

    messages = list(mailbox.fetch())
    assert len(messages) == 1
    m_flagged = messages[0]
    assert Flag.FLAGGED.imap in m_flagged.flags

    assert m.uid == m_flagged.uid

    unflagAction.apply(mailbox, m_flagged, False)
    messages = list(mailbox.fetch())
    assert len(messages) == 1
    m_unflagged = messages[0]
    assert Flag.FLAGGED.imap not in m_unflagged.flags
